from __future__ import annotations

import os
import subprocess
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, runtime_checkable

from pydantic import BaseModel

from autopub.exceptions import AutopubException, CommandFailed
from autopub.types import ReleaseInfo

if TYPE_CHECKING:
    from autopub import ConfigType

Config = TypeVar("Config", bound=BaseModel)


class AutopubPlugin:
    id: str
    data: dict[str, object] = {}

    _config: ConfigType | None = None

    def validate_config(self, config: ConfigType):
        configuration_class: type[BaseModel] | None = getattr(self, "Config", None)

        if configuration_class is None:
            return

        # Support both tool.autopub.plugins.<id> and tool.autopub.plugin_config.<id>
        plugins_value = config.get("plugins", {})

        # plugins can be a list of plugin module paths (e.g. ["autopub.plugins.github"])
        # or a dict of plugin configs (e.g. {"git": {...}}). Only use .get() if it's a dict.
        if isinstance(plugins_value, Mapping):
            plugin_config = plugins_value.get(self.id, {})
        else:
            plugin_config = {}

        if not plugin_config:
            # Fallback to legacy plugin_config location
            plugin_config = config.get("plugin_config", {}).get(self.id, {})

        self._config = configuration_class.model_validate(plugin_config)

    @property
    def config(self) -> ConfigType:
        assert self._config is not None
        return self._config

    def run_command(self, command: list[str]) -> None:
        try:
            subprocess.run(command, check=True, env=os.environ.copy())
        except subprocess.CalledProcessError as e:
            raise CommandFailed(command=command, returncode=e.returncode) from e

    def post_check(self, release_info: ReleaseInfo) -> None:  # pragma: no cover
        ...

    def prepare(self, release_info: ReleaseInfo) -> None:  # pragma: no cover
        ...

    def post_prepare(self, release_info: ReleaseInfo) -> None:  # pragma: no cover
        ...

    def validate_release_notes(
        self, release_info: ReleaseInfo
    ) -> None:  # pragma: no cover
        ...

    def on_release_notes_valid(
        self, release_info: ReleaseInfo
    ) -> None:  # pragma: no cover
        ...

    def on_release_file_not_found(self) -> None:  # pragma: no cover
        ...

    def on_release_notes_invalid(
        self, exception: AutopubException
    ) -> None:  # pragma: no cover
        ...

    def post_publish(self, release_info: ReleaseInfo) -> None:  # pragma: no cover
        ...


@runtime_checkable
class AutopubPackageManagerPlugin(Protocol):
    def build(self) -> None:  # pragma: no cover
        ...

    def publish(
        self, repository: str | None = None, **kwargs: Any
    ) -> None:  # pragma: no cover
        ...



from __future__ import annotations

import os

from pydantic import BaseModel, Field

from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo

COMMIT_TEMPLATE = """\
🤖 Release {release_info.version}

{release_info.release_notes}

[skip ci]
"""


class GitConfig(BaseModel):
    """Git configuration for autopub."""

    git_username: str = Field(
        default="autopub",
        description="Git username for commits",
        validation_alias="git-username",
    )
    git_email: str = Field(
        default="autopub@autopub",
        description="Git email for commits",
        validation_alias="git-email",
    )


class GitPlugin(AutopubPlugin):
    id = "git"
    Config = GitConfig

    def _get_git_username(self) -> str:
        """Get git username from environment variable or config."""
        env_value = os.environ.get("GIT_USERNAME")
        if env_value:
            return env_value

        if self._config:
            return self.config.git_username  # type: ignore

        return "autopub"

    def _get_git_email(self) -> str:
        """Get git email from environment variable or config."""
        env_value = os.environ.get("GIT_EMAIL")
        if env_value:
            return env_value

        if self._config:
            return self.config.git_email  # type: ignore

        return "autopub@autopub"

    def post_publish(self, release_info: ReleaseInfo) -> None:
        assert release_info.version is not None

        tag_name = release_info.version

        git_email = self._get_git_email()
        git_username = self._get_git_username()

        self.run_command(["git", "config", "--global", "user.email", git_email])
        self.run_command(["git", "config", "--global", "user.name", git_username])

        self.run_command(["git", "tag", tag_name])

        commit_message = COMMIT_TEMPLATE.format(release_info=release_info)

        # TODO: config?
        self.run_command(["git", "rm", "RELEASE.md"])
        # TODO: this fails if autopub is git-ignored
        self.run_command(["git", "add", "--all", "--", ":!.autopub"])
        self.run_command(["git", "commit", "-m", commit_message])
        self.run_command(["git", "push"])
        self.run_command(["git", "push", "origin", tag_name])



from __future__ import annotations

import pathlib
import re

import tomlkit
from dunamai import Version

from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo


class BumpVersionPlugin(AutopubPlugin):
    @property
    def pyproject_config(self) -> tomlkit.TOMLDocument:
        content = pathlib.Path("pyproject.toml").read_text()

        return tomlkit.parse(content)

    def _get_version(self, config: tomlkit.TOMLDocument) -> str:
        try:
            return config["tool"]["poetry"]["version"]  # type: ignore
        except KeyError:
            return config["project"]["version"]  # type: ignore

    def _update_version(self, config: tomlkit.TOMLDocument, new_version: str) -> None:
        try:
            config["tool"]["poetry"]["version"] = new_version  # type: ignore
        except KeyError:
            config["project"]["version"] = new_version  # type: ignore

    def post_check(self, release_info: ReleaseInfo) -> None:
        config = self.pyproject_config

        bump_type = {"major": 0, "minor": 1, "patch": 2}[release_info.release_type]

        version = Version(self._get_version(config))

        release_info.previous_version = str(version)
        release_info.version = version.bump(bump_type).serialize()

    def _get_package_name(self, config: tomlkit.TOMLDocument) -> str | None:
        """Get the package name from pyproject.toml."""
        try:
            return config["tool"]["poetry"]["name"]  # type: ignore
        except KeyError:
            try:
                return config["project"]["name"]  # type: ignore
            except KeyError:
                return None

    def _find_package_init(self, package_name: str) -> pathlib.Path | None:
        """Find the package's __init__.py file.

        Tries multiple common layouts:
        - src/package_name/__init__.py
        - package_name/__init__.py
        - src/__init__.py (for single-module packages)
        """
        # Convert package name with hyphens to underscores for directory name
        package_dir = package_name.replace("-", "_")

        possible_paths = [
            pathlib.Path("src") / package_dir / "__init__.py",
            pathlib.Path(package_dir) / "__init__.py",
            pathlib.Path("src") / "__init__.py",
        ]

        for path in possible_paths:
            if path.exists():
                return path

        return None

    def _update_init_version(self, new_version: str) -> None:
        """Update __version__ in the package's __init__.py file if it exists."""
        config = self.pyproject_config
        package_name = self._get_package_name(config)

        if not package_name:
            return

        init_file = self._find_package_init(package_name)

        if not init_file:
            return

        content = init_file.read_text()

        # Regex pattern to match __version__ = "x.y.z" or __version__ = 'x.y.z'
        pattern = r'__version__\s*=\s*["\'][\d.]+["\']'

        # Check if __version__ exists in the file
        if not re.search(pattern, content):
            return

        # Replace the version
        new_content = re.sub(pattern, f'__version__ = "{new_version}"', content)

        init_file.write_text(new_content)

    def post_prepare(self, release_info: ReleaseInfo) -> None:
        config = self.pyproject_config

        assert release_info.version is not None

        self._update_version(config, release_info.version)

        pathlib.Path("pyproject.toml").write_text(tomlkit.dumps(config))  # type: ignore

        # Update __version__ in __init__.py if it exists
        self._update_init_version(release_info.version)



from __future__ import annotations

from typing import Any

from autopub.plugins import AutopubPackageManagerPlugin
from autopub.plugins.bump_version import BumpVersionPlugin

__all__ = ["PDMPlugin"]


class PDMPlugin(BumpVersionPlugin, AutopubPackageManagerPlugin):
    def build(self) -> None:
        self.run_command(["pdm", "build"])

    def publish(self, repository: str | None = None, **kwargs: Any) -> None:
        additional_args: list[str] = []

        if repository:
            additional_args += ["--repository", repository]

        self.run_command(["pdm", "publish", *additional_args])



from __future__ import annotations

from datetime import date
from pathlib import Path

from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo

# TODO: from config
CHANGELOG_HEADER = "========="
VERSION_HEADER = "-"


class UpdateChangelogPlugin(AutopubPlugin):
    @property
    def changelog_file(self) -> Path:
        return Path("CHANGELOG.md")

    def post_prepare(self, release_info: ReleaseInfo) -> None:
        if not self.changelog_file.exists():
            self.changelog_file.write_text(f"CHANGELOG\n{CHANGELOG_HEADER}\n\n")

        current_date = date.today().strftime("%Y-%m-%d")

        lines = self.changelog_file.read_text().splitlines()

        # Look for CHANGELOG header (e.g., "CHANGELOG\n=========")
        header = []
        old_changelog_data = []
        header_found = False

        for index, line in enumerate(lines):
            if CHANGELOG_HEADER == line.strip():
                old_changelog_data = lines[index + 1 :]
                header = lines[: index + 1]
                header_found = True
                break

        # If no header found, treat all existing content as old changelog data
        if not header_found:
            old_changelog_data = lines

        new_version = release_info.version

        assert new_version is not None

        with self.changelog_file.open("w") as f:
            # Write header if it exists
            if header:
                f.write("\n".join(header))
                f.write("\n\n")

            # Write new version entry
            new_version_header = f"{new_version} - {current_date}"
            f.write(f"{new_version_header}\n")
            f.write(f"{VERSION_HEADER * len(new_version_header)}\n\n")
            f.write(release_info.release_notes)

            if release_info.additional_release_notes:
                for line in release_info.additional_release_notes:
                    f.write(f"\n\n{line}")

            # Write old changelog data (skip if empty or only whitespace)
            old_content = "\n".join(old_changelog_data).strip()
            if old_content:
                f.write("\n\n")
                # Strip leading empty lines from old changelog data to avoid extra blank lines
                while old_changelog_data and old_changelog_data[0].strip() == "":
                    old_changelog_data = old_changelog_data[1:]
                # Preserve the original formatting including any trailing newlines
                f.write("\n".join(old_changelog_data))



from typing import Annotated, Optional, TypedDict

import rich
import typer
from rich.console import Group, RenderableType
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel

from autopub import Autopub
from autopub.exceptions import AutopubException, InvalidConfiguration

app = typer.Typer()


class State(TypedDict):
    plugins: list[str]


class AutoPubCLI(typer.Context):
    obj: Autopub


@app.command()
def check(context: AutoPubCLI):
    """This commands checks if the current PR has a valid release file."""

    autopub = context.obj

    try:
        autopub.check()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        release_info = autopub.release_info

        rich.print(
            Padding(
                Group(
                    (
                        "[bold on bright_magenta] Release type: [/] "
                        f"[yellow italic underline]{release_info.release_type}[/]\n"
                    ),
                    "[bold on bright_magenta] Release notes: [/]\n",
                    Markdown(release_info.release_notes),
                    "\n---\n\n[green bold]Release file is valid![/] 🚀",
                ),
                (1, 1),
            )
        )


@app.command()
def build(context: AutoPubCLI):
    autopub = context.obj

    try:
        autopub.build()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Build succeeded"))


@app.command()
def prepare(context: AutoPubCLI):
    autopub = context.obj

    try:
        autopub.prepare()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Preparation succeeded"))


@app.command()
def publish(
    context: AutoPubCLI,
    repository: Annotated[
        Optional[str],
        typer.Option("--repository", "-r", help="Repository to publish to"),
    ] = None,
):
    autopub = context.obj

    try:
        autopub.publish(repository=repository)
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Publishing succeeded"))


@app.callback(invoke_without_command=True)
def main(
    context: AutoPubCLI,
    should_show_version: Annotated[
        Optional[bool], typer.Option("--version", is_eager=True)
    ] = None,
):
    if should_show_version:
        from importlib.metadata import version

        print(version("autopub"))

        raise typer.Exit()

    autopub = Autopub()

    # default plugins we always want to load (?)
    autopub.load_plugins(["git", "update_changelog", "bump_version"])

    try:
        autopub.validate_config()
    except InvalidConfiguration as e:
        title = "[red]🚨 Some of the plugins have invalid configuration[/]"

        parts: list[RenderableType] = []

        for id_ in e.validation_errors:
            error = e.validation_errors[id_]
            parts.append("")
            parts.append(f"[bold on bright_magenta] Plugin: [/] {id_}")
            parts.append("")

            errors: list[RenderableType] = []

            for error in error.errors():
                location = " -> ".join(map(str, error["loc"]))
                message = error["msg"]

                errors.append(f"[bold on blue] {location} [/]: {message}")
                errors.append("")

            parts.append(Padding(Group(*errors), (0, 2)))

        content = Group(f"[red]{title}[/]", *parts)

        rich.print(Padding(content, (1, 1)))

        raise typer.Exit(1) from e

    context.obj = autopub
